#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/stat.h>

#include "lib.h"

int delete(char *filename, int recursive) {
	DIR *dir;
	struct stat st;
	struct dirent *de;
	char *s;
	int result;
	result = 0;
	if (lstat(filename, &st))
		return -1;
	if (S_ISDIR(st.st_mode)) {
		if (recursive) {
			dir = opendir(filename);
			if (!dir)
				return -1;
			for (de = readdir(dir); de; de = readdir(dir)) {
				if (strcmp(de->d_name, ".") && strcmp(de->d_name, "..")) {
					s = malloc(strlen(filename) + strlen(de->d_name) + 2);
					if (!s) {
						errno = ENOMEM;
						return -1;
					}
					sprintf(s, "%s/%s", filename, de->d_name);
					if (delete(s, 1))
						result = -1;
					free(s);
				}
			}
		}
		if (rmdir(filename))
			result = -1;
		return result;
	} else {
		return unlink(filename);
	}
}


void compress_file(char *filename, char *new,
		struct stream *(*dopen_function)(int, char *, int), int level,
		int force) {
	struct stream *in, *out;
	struct stat st;
	int fd, ch, err, flags;
	if (strcmp(filename, "-") && stat(filename, &st)) {
		error(filename);
		return;
	}
	in = file_stream_open(filename, "r");
	if (!in)
		return;
	if (strcmp(new, "-")) {
		flags = O_WRONLY|O_CREAT|O_TRUNC;
		if (!force)
			flags |= O_EXCL;
		fd = open(new, flags, st.st_mode);
	} else {
		fd = dup(1);
	}
	if (fd < 0) {
		error(new);
		in->close(in);
		return;
	}
	out = dopen_function(fd, new, level);
	if (!out) {
		in->close(in);
		close(fd);
		delete(new, 0);
		return;
	}
	copy_all(in, out);
	err = in->error(in) || out->error(out);
	out->close(out);
	in->close(in);
	if (err)
		delete(new, 0);
	else
		delete(filename, 0);
}

void decompress(char *filename, char *new,
		struct stream *(*open_function)(char *, char *), int force) {
	struct stream *in, *out;
	struct stat st;
	FILE *f;
	int fd, ch, err, flags;
	if (strcmp(filename, "-") && stat(filename, &st)) {
		error(filename);
		return;
	}
	in = open_function(filename, "r");
	if (!in)
		return;
	if (strcmp(new, "-")) {
		flags = O_WRONLY|O_CREAT|O_TRUNC;
		if (!force)
			flags |= O_EXCL;
		fd = open(new, flags, st.st_mode);
	} else {
		fd = dup(1);
	}
	if (fd < 0) {
		error(new);
		in->close(in);
		return;
	}
	f = fdopen(fd, "w");
	if (!f) {
		error(new);
		close(fd);
		in->close(in);
		return;
	}
	out = file_stream_new(f, new);
	if (!out) {
		error(new);
		fclose(f);
		in->close(in);
		return;
	}
	copy_all(in, out);
	err = in->error(in) || out->error(out);
	out->close(out);
	in->close(in);
	if (err)
		delete(new, 0);
	else
		delete(filename, 0);
}

void version(char *cmd) {
	fprintf(stderr, "Tuesday Software Package Manager Version 3: %s\n", cmd);
}

void gz_compress_file_function(char *filename, int level, int force) {
	char *new;
	size_t len;
	if (!strcmp(filename, "-")) {
		compress_file("-", "-", gz_compress_dopen, level, force);
	} else {
		len = strlen(filename);
		new = emalloc(len + 4, filename);
		if (new) {
			sprintf(new, "%s.gz", filename);
			compress_file(filename, new, gz_compress_dopen, level, force);
			free(new);
		}
	}
}

void gz_decompress_file_function(char *filename, int force) {
	char *new;
	size_t len;
	if (!strcmp(filename, "-")) {
		decompress("-", "-", gz_stream_open, force);
	} else {
		len = strlen(filename);
		if (len > 3 && !strcmp(filename + len - 3, ".gz")) {
			new = emalloc(len - 2, filename);
			if (new) {
				strncpy(new, filename, len - 3);
				new[len - 3] = 0;
				decompress(filename, new, gz_stream_open, force);
				free(new);
			}
		} else if (len > 4 && !strcmp(filename + len - 4, ".tgz")) {
			new = emalloc(len + 1, filename);
			if (new) {
				strncpy(new, filename, len - 2);
				strncpy(new + len - 2, "ar", 2);
				new[len] = 0;
				decompress(filename, new, gz_stream_open, force);
				free(new);
			}
		} else {
			fprintf(stderr, "%s: %s: not a GZIP filename\n", progname,
				filename);
			exit(1);
		}
	}
}

void bz2_compress_file_function(char *filename, int level, int force) {
	char *new;
	size_t len;
	if (!strcmp(filename, "-")) {
		compress_file("-", "-", bz2_compress_dopen, level, force);
	} else {
		len = strlen(filename);
		new = emalloc(len + 5, filename);
		if (new) {
			sprintf(new, "%s.bz2", filename);
			compress_file(filename, new, bz2_compress_dopen, level, force);
			free(new);
		}
	}
}

void bz2_decompress_file_function(char *filename, int force) {
	char *new;
	size_t len;
	if (!strcmp(filename, "-")) {
		decompress("-", "-", bz2_stream_open, force);
	} else {
		len = strlen(filename);
		if (len > 4 && !strcmp(filename + len - 4, ".bz2")) {
			new = emalloc(len - 3, filename);
			if (new) {
				strncpy(new, filename, len - 4);
				new[len - 4] = 0;
				decompress(filename, new, bz2_stream_open, force);
				free(new);
			}
		} else if (len > 5 && !strcmp(filename + len - 5, ".tbz2")) {
			new = emalloc(len, filename);
			if (new) {
				strncpy(new, filename, len - 3);
				strncpy(new + len - 3, "ar", 2);
				new[len - 1] = 0;
				decompress(filename, new, bz2_stream_open, force);
				free(new);
			}
		} else {
			fprintf(stderr, "%s: %s: not a BZIP2 filename\n", progname,
				filename);
			exit(1);
		}
	}
}

void xz_compress_file_function(char *filename, int level, int force) {
	char *new;
	size_t len;
	if (!strcmp(filename, "-")) {
		compress_file("-", "-", xz_compress_dopen, level, force);
	} else {
		len = strlen(filename);
		new = xmalloc(len + 4, filename);
		sprintf(new, "%s.xz", filename);
		compress_file(filename, new, xz_compress_dopen, level, force);
	}
}

void xz_decompress_file_function(char *filename, int force) {
	char *new;
	size_t len;
	if (!strcmp(filename, "-")) {
		decompress("-", "-", xz_stream_open, force);
	} else {
		len = strlen(filename);
		if (len > 3 && !strcmp(filename + len - 3, ".xz")) {
			new = emalloc(len - 2, filename);
			if (new) {
				strncpy(new, filename, len - 3);
				new[len - 3] = 0;
				decompress(filename, new, xz_stream_open, force);
				free(new);
			}
		} else if (len > 4 && !strcmp(filename + len - 4, ".txz")) {
			new = emalloc(len + 1, filename);
			if (new) {
				strncpy(new, filename, len - 2);
				strncpy(new + len - 2, "ar", 2);
				new[len] = 0;
				decompress(filename, new, xz_stream_open, force);
				free(new);
			}
		} else {
			fprintf(stderr, "%s: %s: not an XZ filename\n", progname,
				filename);
			exit(1);
		}
	}
}

int cat_process_file(char *filename, struct stream *out,
		struct stream *(*open_function)(char *, char *)) {
	struct stream *in;
	in = open_function(filename, "r");
	if (in) {
		copy_all(in, out);
		if (out->error(out))
			return 1;
		in->close(in);
	}
	return 0;
}

int cat_command_main(int argc, char **argv,
		struct stream *(*open_function)(char *, char *), char *cmd,
		char *usage) {
	struct stream *out;
	size_t i;
	int result;
	progname = argv[0];
	out = file_stream_new(stdout, "(standard output)");
	if (!out)
		return 1;
	for (i = 1; i < argc && argv[i][0] == '-' && argv[i][1]; i++) {
		if (!strcmp(argv[i], "--version")) {
			version(cmd);
			return 0;
		} else if (!strcmp(argv[i], "--help")) {
			version(cmd);
			fprintf(stderr,
				"\n%s [INPUT ...]\n"
				"\t%s\n"
				"%s --version\n"
				"\tDisplay the version of this command.\n"
				"%s --help\n"
				"\tDisplay this usage message.\n",
				progname, usage, progname, progname);
			return 0;
		} else {
			fprintf(stderr, "%s: invalid %s (%s --help for usage)\n", progname,
				argv[i], progname);
			return 1;
		}
	}
	result = 0;
	if (i == argc && cat_process_file("-", out, open_function))
		result = 1;
	for (; i < argc; i++)
		if (cat_process_file(argv[i], out, open_function))
			result = 1;
	return result;
}

int unzip_command_main(int argc, char **argv,
		void (*decompress_file_function)(char *, int), char *cmd,
		char *usage) {
	int i, j, force;
	progname = argv[0];
	force = 0;
	for (i = 1; i < argc && argv[i][0] == '-' && argv[i][1]; i++) {
		if (!strcmp(argv[i], "--version")) {
			version(cmd);
			return 0;
		} else if (!strcmp(argv[i], "--help")) {
			version(cmd);
			fprintf(stderr,
				"\n%s [INPUT ...]\n"
				"\t%s\n"
				"%s --version\n"
				"\tDisplay the version of this command.\n"
				"%s --help\n"
				"\tDisplay this usage message.\n"
				"%s -f\n"
				"\tOverwrite the output file.\n",
				progname, usage, progname, progname, progname);
			return 0;
		} else if (argv[i][1] != '-') {
			for (j = 1; argv[i][j]; j++) {
				switch (argv[i][j]) {
				case 'f':
					force = 1;
					break;
				default:
					fprintf(stderr, "%s: invalid -%s (%s --help for usage)\n",
						progname, argv[i][j], progname);
					return 1;
				}
			}
		} else {
			fprintf(stderr, "%s: invalid %s (%s --help for usage)\n", progname,
				argv[i], progname);
			return 1;
		}
	}
	if (i >= argc)
		decompress_file_function("-", force);
	for (; i < argc; i++)
		decompress_file_function(argv[i], force);
	return 0;
}

int zip_command_main(int argc, char **argv,
		void (*compress_file_function)(char *, int, int),
		void (*decompress_file_function)(char *, int), char *cmd, char *usage,
		int min_level) {
	int i, j, level, force, reverse;
	char *new;
	progname = argv[0];
	level = 6;
	force = 0;
	reverse = 0;
	for (i = 1; i < argc && argv[i][0] == '-' && argv[i][1]; i++) {
		if (!strcmp(argv[i], "--version")) {
			version(cmd);
			return 0;
		} else if (!strcmp(argv[i], "--help")) {
			version(cmd);
			fprintf(stderr,
				"\n%s [INPUT ...]\n"
				"\t%s\n"
				"%s --version\n"
				"\tDisplay the version of this command.\n"
				"%s --help\n"
				"\tDisplay this usage message.\n"
				"%s -d\n"
				"\tDecompress rather than compress.\n"
				"%s -f\n"
				"\tOverwrite the output file.\n"
				"%s -n\n"
				"\tDo not save a filename in the compressed output. This is\n"
				"\tthe default; the option is for compatibility.\n"
				"%s -#\n"
				"\tWhere # is a digit %d through 9, set compression level\n"
				"\tto #.\n",
				progname, usage, progname, progname, progname, progname,
				progname, progname, min_level);
			return 0;
		} else if (argv[i][1] != '-') {
			for (j = 1; argv[i][j]; j++) {
				switch (argv[i][j]) {
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
					level = argv[i][j] - '0';
					if (level >= min_level)
						break;
				case 'd':
					reverse = 1;
					break;
				case 'f':
					force = 1;
				case 'n':
					break;
				default:
					fprintf(stderr, "%s: invalid -%s (%s --help for usage)\n",
						progname, argv[i][j], progname);
					return 1;
				}
			}
		} else {
			fprintf(stderr, "%s: invalid %s (%s --help for usage)\n", progname,
				argv[i], progname);
			return 1;
		}
	}
	if (i >= argc) {
		if (reverse)
			decompress_file_function("-", force);
		else
			compress_file_function("-", level, force);
	}
	for (; i < argc; i++)
		if (reverse)
			decompress_file_function(argv[i], force);
		else
			compress_file_function(argv[i], level, force);
	return 0;
}
